/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb.server;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.net.ssl.SSLContext;



import org.xlightweb.HttpUtils;
import org.xlightweb.IHttpConnectionHandler;
import org.xlightweb.IHttpRequestHandler;
import org.xlightweb.server.TransactionMonitor.Transaction;
import org.xlightweb.server.TransactionMonitor.TransactionLog;
import org.xsocket.connection.ConnectionUtils;
import org.xsocket.connection.Server;


 


/**
 * A HttpServer. The http server accepts incoming connections and forwards the request to
 * the assigned {@link IHttpRequestHandler}. Example:
 * 
 * <pre>
 * 
 *  // defining a request handler 
 *  class MyRequestHandler implements IHttpRequestHandler  {
 *  
 *     public void onRequest(IHttpExchange exchange) throws IOException {
 *        IHttpRequest request = exchange.getRequest();
 *        //...
 *        
 *        exchange.send(new HttpResponse(200, "text/plain", "it works"));
 *     }
 *  } 
 *
 * 
 *  // creates a server
 *  IServer server = new HttpServer(8080, new MyRequestHandler());
 * 
 *  // setting some properties 
 *  server.setMaxTransactions(400);
 *  server.setRequestTimeoutMillis(5 * 60 * 1000);  
 *  //...
 * 
 * 
 *  // executing the server 
 *  server.run();  // (blocks forever)
 * 
 *  // or server.start();  (blocks until the server has been started)
 *  //...
 * </pre> 
 *
 * 
 * @author grro@xlightweb.org
 */
public class HttpServer extends Server implements IHttpServer {

    

    // transaction monitor
    private TransactionMonitor transactionMonitor = null;
    private final TransactionLog transactionLog = new TransactionLog(0);


    
    
	/**
	 * constructor <br><br>
	 *
	 * @param requestHandler the requestHandler
	 * @throws IOException If some other I/O error occurs
	 * @throws UnknownHostException if the local host cannot determined
	 */
	public HttpServer(IHttpRequestHandler requestHandler) throws UnknownHostException, IOException {
		this(new InetSocketAddress(0), new HashMap<String, Object>(), requestHandler, null, false);
	}



	/**
	 * constructor <br><br>
	 *
	 * @param handler              the handler to use
	 * @param options              the socket options  
	 * @throws IOException If some other I/O error occurs
	 * @throws UnknownHostException if the local host cannot determined
	 */
	public HttpServer(IHttpRequestHandler handler, Map<String, Object> options) throws UnknownHostException, IOException {
		this(new InetSocketAddress(0), options, handler, null, false);
	}


	/**
	 * constructor  <br><br>
	 *
	 * @param port        the local port
	 * @param handler     the handler to use
	 * @throws UnknownHostException if the local host cannot determined
     * @throws IOException If some other I/O error occurs
	 */
	public HttpServer(int port, IHttpRequestHandler handler) throws UnknownHostException, IOException {
		this(new InetSocketAddress(port), new HashMap<String, Object>(), handler, null, false);
	}




	/**
	 * constructor <br><br>
	 *
	 *
	 * @param port                 the local port
	 * @param handler              the handler to use
	 * @param options              the acceptor socket options 
	 * @throws UnknownHostException if the local host cannot determined
     * @throws IOException If some other I/O error occurs
	 */
	public HttpServer(int port, IHttpRequestHandler handler, Map<String , Object> options) throws UnknownHostException, IOException {
		this(new InetSocketAddress(port), options, handler, null, false);
	}


	/**
	 * constructor <br><br>
	 *
	 *
	 * @param address     the local address
	 * @param port        the local port
	 * @param handler     the handler to use
	 * @throws UnknownHostException if the local host cannot determined
     * @throws IOException If some other I/O error occurs
	 */
	public HttpServer(InetAddress address, int port, IHttpRequestHandler handler) throws UnknownHostException, IOException {
		this(address, port, handler, new HashMap<String, Object>(), null, false);
	}



	/**
	 * constructor  <br><br>
	 *
	 *
	 * @param ipAddress   the local ip address
	 * @param port        the local port
	 * @param handler     the handler to use 
	 * @throws UnknownHostException if the local host cannot determined
     * @throws IOException If some other I/O error occurs
	 */
	public HttpServer(String ipAddress, int port, IHttpRequestHandler handler) throws UnknownHostException, IOException {
		this(InetAddress.getByName(ipAddress), port, handler, new HashMap<String, Object>(), null, false);
	}



	/**
	 * constructor <br><br>
	 *
	 *
	 * @param ipAddress            the local ip address
	 * @param port                 the local port
	 * @param handler              the handler to use 
	 * @param options              the socket options 
	 * @throws UnknownHostException if the local host cannot determined
     * @throws IOException If some other I/O error occurs
	 */
	public HttpServer(String ipAddress, int port, IHttpRequestHandler handler, Map<String, Object> options) throws UnknownHostException, IOException {
		this(InetAddress.getByName(ipAddress), port, handler, options, null, false);
	}




	/**
	 * constructor <br><br>
	 *
	 *
	 * @param port               local port
	 * @param handler            the handler to use
	 * @param sslOn              true, is SSL should be activated
	 * @param sslContext         the ssl context to use
	 * @throws UnknownHostException if the local host cannot determined
	 * @throws IOException If some other I/O error occurs
	 */
	public HttpServer(int port, IHttpRequestHandler handler, SSLContext sslContext, boolean sslOn) throws UnknownHostException, IOException {
		this(new InetSocketAddress(port), new HashMap<String, Object>(), handler, sslContext, sslOn);
	}



	/**
	 * constructor <br><br>
	 *
	 * @param port                 local port
	 * @param options              the acceptor socket options
	 * @param handler              the handler to use 
	 * @param sslOn                true, is SSL should be activated
	 * @param sslContext           the ssl context to use
	 * @throws UnknownHostException if the local host cannot determined
	 * @throws IOException If some other I/O error occurs
	 */
	public HttpServer(int port, IHttpRequestHandler handler, Map<String, Object> options, SSLContext sslContext, boolean sslOn) throws UnknownHostException, IOException {
		this(new InetSocketAddress(port), options, handler, sslContext, sslOn);
	}


	/**
	 * constructor <br><br>
	 *
	 * @param ipAddress          local ip address
	 * @param port               local port
	 * @param handler            the handler to use 
	 * @param sslOn              true, is SSL should be activated
	 * @param sslContext         the ssl context to use
	 * @throws UnknownHostException if the local host cannot determined
	 * @throws IOException If some other I/O error occurs
	 */
	public HttpServer(String ipAddress, int port, IHttpRequestHandler handler, SSLContext sslContext, boolean sslOn) throws UnknownHostException, IOException {
		this(InetAddress.getByName(ipAddress), port, handler, new HashMap<String, Object>(),  sslContext, sslOn);
	}



	/**
	 * constructor <br><br>
	 *
	 *
	 * @param ipAddress            local ip address
	 * @param port                 local port
	 * @param options              the acceptor socket options
	 * @param handler              the handler to use 
	 * @param sslOn                true, is SSL should be activated
	 * @param sslContext           the ssl context to use
	 * @throws UnknownHostException if the local host cannot determined
	 * @throws IOException If some other I/O error occurs
	 */
	public HttpServer(String ipAddress, int port, IHttpRequestHandler handler, Map<String, Object> options, SSLContext sslContext, boolean sslOn) throws UnknownHostException, IOException {
		this(InetAddress.getByName(ipAddress), port, handler, options, sslContext, sslOn);
	}


	/**
	 * constructor <br><br>
	 *
	 *
	 * @param address            local address
	 * @param port               local port
	 * @param handler            the request handler
	 * @param sslOn              true, is SSL should be activated
	 * @param sslContext         the ssl context to use
	 * @throws UnknownHostException if the local host cannot determined
	 * @throws IOException If some other I/O error occurs
	 */
	public HttpServer(InetAddress address, int port, IHttpRequestHandler handler, SSLContext sslContext, boolean sslOn) throws UnknownHostException, IOException {
		this(address, port, handler, new HashMap<String, Object>(), sslContext, sslOn);
	}




	/**
	 * constructor <br><br>
	 *
	 *
	 * @param address              local address
	 * @param port                 local port
	 * @param options              the socket options
	 * @param handler              the request handler 
	 * @param sslOn                true, is SSL should be activated
	 * @param sslContext           the ssl context to use
	 * @throws UnknownHostException if the local host cannot determined
	 * @throws IOException If some other I/O error occurs
	 */
	public HttpServer(InetAddress address, int port, IHttpRequestHandler handler, Map<String, Object> options, SSLContext sslContext, boolean sslOn) throws UnknownHostException, IOException {
		this(new InetSocketAddress(address, port), options, handler, sslContext, sslOn);
	}




	private HttpServer(InetSocketAddress address, Map<String, Object> options, IHttpRequestHandler requestHandler, SSLContext sslContext, boolean sslOn) throws UnknownHostException, IOException {
		super(address, options, new HttpProtocolAdapter(requestHandler), sslContext, sslOn, 0);
	}

	
	
	/**
	 * adds a connection handler 
	 * 
	 * @param connectionHandler  the connection handler to add
	 */
	public void addConnectionHandler(IHttpConnectionHandler connectionHandler) {
		((HttpProtocolAdapter) super.getHandler()).addConnectionHandler(connectionHandler);
	}

	
	/**
     * returns the implementation version
     *  
     * @return the implementation version
     */
    public String getImplementationVersion() {
        return HttpUtils.getImplementationVersion();
    }
    
    
    int getNumHandledConnections() {
        return ((HttpProtocolAdapter) super.getHandler()).getNumHandledConnections();
    }
    
    
    /**
     * returns the implementation date
     * 
     * @return  the implementation date 
     */
    public String getImplementationDate() {
        return HttpUtils.getImplementationDate();
    }
	

    /**
     * {@inheritDoc}
     */
    public IHttpRequestHandler getRequestHandler() {
    	return (IHttpRequestHandler) ((HttpProtocolAdapter) getHandler()).getRequestHandler();
    }

    
    /**
     * returns the xSocket implementation version
     *  
     * @return the xSocket implementation version
     */ 
    String getXSocketImplementationVersion() {
        return ConnectionUtils.getImplementationVersion();
    }
    
    
    
    /**
     * returns the xScoket implementation date
     * 
     * @return  the xSocket implementation date 
     */
    String getXSocketImplementationDate() {
        return ConnectionUtils.getImplementationDate();
    }
	
	/**
	 * sets the message receive timeout
	 * 
	 * @param receivetimeout the message receive timeout
	 */
	public void setRequestTimeoutMillis(long receivetimeout) {
		((HttpProtocolAdapter) super.getHandler()).setRequestTimeoutMillis(receivetimeout);
	}

		
	   
    /**
     * gets the message receive timeout
     * 
     * @return the message receive timeout
     */
    public long getRequestTimeoutMillis() {
        return ((HttpProtocolAdapter) super.getHandler()).getRequestTimeoutMillis();
    }
    
    
	
	/**
	 * set the body data receive timeout
	 * 
	 * @param bodyDataReceiveTimeoutSec the timeout
	 */
	public final void setBodyDataReceiveTimeoutMillis(long bodyDataReceiveTimeoutMillis) {
		((HttpProtocolAdapter) super.getHandler()).setBodyDataReceiveTimeoutMillis(bodyDataReceiveTimeoutMillis);
	}


   /**
     * sets the session max inactive interval in seconds
     *  
     * @param sessionMaxInactiveIntervalSec the session max inactive interval in seconds
     */
    public void setSessionMaxInactiveIntervalSec(int sessionMaxInactiveIntervalSec) {
        ((HttpProtocolAdapter) super.getHandler()).setSessionMaxInactiveIntervalSec(sessionMaxInactiveIntervalSec);
    }

    
    /**
     * gets the session max inactive interval in sconds
     * 
     * @return the session max inactive interval in sconds
     */
    public int getSessionMaxInactiveIntervalSec() {
        return ((HttpProtocolAdapter) super.getHandler()).getSessionMaxInactiveIntervalSec();
    }
	
	
	/**
	 * set is if the server-side connection will closed, if an error message is sent
	 * 
	 * @param isCloseOnSendingError if the connection will closed, if an error message is sent
	 */
	public void setCloseOnSendingError(boolean isCloseOnSendingError) {
		((HttpProtocolAdapter) super.getHandler()).setCloseOnSendingError(isCloseOnSendingError);
	}
	
	
	/**
	 * returns if the server-side connection will closed, if an error message will be sent
	 * @return true, if the connection will closed by sending an error message 
	 */
	public boolean isCloseOnSendingError() {
		return ((HttpProtocolAdapter) super.getHandler()).isCloseOnSendingError();
	}
	
	
	
	/**
	 * set the max transactions per connection. Setting this filed causes that 
	 * a keep-alive response header will be added 
	 * 
	 * @param maxTransactions  the max transactions 
	 */
	public void setMaxTransactions(int maxTransactions) {
		((HttpProtocolAdapter) getHandler()).setMaxTransactions(maxTransactions);
	}

	   
    /**
     * get the max transactions per connection. Setting this filed causes that 
     * a keep-alive response header will be added 
     * 
     * @return the max transactions 
     */
    public int getMaxTransactions() {
        return ((HttpProtocolAdapter) getHandler()).getMaxTransactions();
    }
    
	
	/**
	 * sets the session manager
	 * 
	 * @param sessionManager the session manager
	 */
	public void setSessionManager(ISessionManager sessionManager) {
		((HttpProtocolAdapter) getHandler()).setSessionManager(sessionManager);
	}
	
	

	/**
	 * sets if cookies is used for session state management 
	 *  
	 * @param useCookies true, if cookies isused for session state management 
	 */
	public void setUsingCookies(boolean useCookies) {
		((HttpProtocolAdapter) getHandler()).setUsingCookies(useCookies);
	}
	

	/**
	 * returns true, if cookies is used for session state management
	 * 
	 * @return true, if cookies is used for session state management
	 */
	public boolean isUsingCookies() {
		return ((HttpProtocolAdapter) getHandler()).isUsingCookies();
	}
	
	/**
	 * returns the session manager
	 * 
	 * @return the session manager
	 */
	public ISessionManager getSessionManager() {
		return ((HttpProtocolAdapter) getHandler()).getSessionManager();
	}
	
	
	/**
     * returns the transaction log 
     * @return the transaction log
     */
    List<String> getTransactionInfos() {
        List<String> result = new ArrayList<String>();
        for (Transaction transaction : transactionLog.getTransactions()) {
            result.add(transaction.toString());
        }
        return result;
    }
    
    
    

    /**
     * sets the max size of the transaction log
     * 
     * @param maxSize the max size of the transaction log
     */
    void setTransactionLogMaxSize(int maxSize) {
       transactionLog.setMaxSize(maxSize);
        
        if (maxSize == 0) {
            transactionMonitor = null;
        } else {
            transactionMonitor = new TransactionMonitor(transactionLog);
        }
        
        ((HttpProtocolAdapter) getHandler()).setTransactionMonitor(transactionMonitor);
    }
    
    Integer getTransactionsPending() {
        if (transactionMonitor != null) {
            return transactionMonitor.getPendingTransactions();
        } else {
            return null;
        }
    }

    

    /**
     * get the max size of the transaction log
     * 
     * @return the max size of the transaction log
     */
    int getTransactionLogMaxSize() {
        return transactionLog.getMaxSize();
    }    
}
